static const char *RcsId = "";
//+=============================================================================
//
// file :         MenloLFC.cpp
//
// description :  C++ source for the MenloLFC and its commands. 
//                The class is derived from Device. It represents the
//                CORBA servant object which will be accessed from the
//                network. All commands which can be executed on the
//                MenloLFC are implemented in this file.
//
// project :      TANGO Device Server
//
// Elettra-Sincrotrone Trieste S.C.p.A. di interesse nazionale
// Strada Statale 14 - km 163,5 in AREA Science Park
// 34149 Basovizza, Trieste ITALY
//
//-=============================================================================
//
//  		This file is generated by POGO
//	(Program Obviously used to Generate tango Object)
//
//         (c) - Software Engineering Group - ESRF
//=============================================================================



//===================================================================
//
//	The following table gives the correspondence
//	between commands and method name.
//
//  Command name   |  Method name
//	----------------------------------------
//  State          |  dev_state()
//  Status         |  dev_status()
//  DeviceAddress  |  device_address()
//  DeviceID       |  device_id()
//  DeviceHWver    |  device_hwver()
//  DeviceFWver    |  device_fwver()
//  DeviceSerial   |  device_serial()
//
//===================================================================


#include <tango.h>
#include <MenloLFC.h>
#include <MenloLFCClass.h>
//#include <Serial.h>	//for SL_RAW, SL_NCHAR, SL_LINE, SL_RETRY
#define SL_RAW        0     /* raw read/write mode */
#define SL_NCHAR      1     /* character read/write mode */
#define SL_LINE       2     /* line read mode */
#define SL_RETRY      3     /* retry read mode */
#include <errno.h>
#include <math.h>	//for ceil
#include <iomanip>      //used for debug dump of db

#include "readthread.h"

namespace MenloLFC_ns
{

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::MenloLFC(string &s)
// 
// description : 	constructor for simulated MenloLFC
//
// in : - cl : Pointer to the DeviceClass object
//      - s : Device name 
//
//-----------------------------------------------------------------------------
MenloLFC::MenloLFC(Tango::DeviceClass *cl,string &s)
:TANGO_BASE_CLASS(cl,s.c_str())
{
	init_device();
}

MenloLFC::MenloLFC(Tango::DeviceClass *cl,const char *s)
:TANGO_BASE_CLASS(cl,s)
{
	init_device();
}

MenloLFC::MenloLFC(Tango::DeviceClass *cl,const char *s,const char *d)
:TANGO_BASE_CLASS(cl,s,d)
{
	init_device();
}
//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::delete_device()
// 
// description : 	will be called at device destruction or at init command.
//
//-----------------------------------------------------------------------------
void MenloLFC::delete_device()
{
	//	Delete device allocated objects
	abortflag = true;
	sleep(1);
	delete mutex;
	if(moxa)
		delete devserial;

	delete menlo;
	delete attr_ISET_read;
	delete attr_IACT_read;
	delete attr_PACT_read;
	delete attr_DIODE_TSET_read;
	delete attr_DIODE_TACT_read;
	delete attr_MODULE_TACT_read;
	delete attr_TCORR_read;
	delete [] attr_TC0_ADJ_ISET_read;
	delete [] attr_TC0_ADJ_TSET_read;
	delete [] attr_TC0_ADJ_IACT_read;
	delete [] attr_TM0_ADJ_TACT_read;
	delete [] attr_TM0_ADJ_PACT_read;
	delete attr_Amplifier_State_read;
	delete [] attr_Snapshot_read;
	
	delete [] attr_Date_read;
	delete [] attr_Time_read;
	delete attr_RTstamp_read;
	delete attr_Tstamp_read;
	delete attr_TstampDiff_read;
	delete attr_TstampStart_read;
	delete attr_Uptime_read;
	
	delete [] attr_ADJ_PACT_1_read;
	delete attr_PACT_1_read;
	delete attr_PMIN_1_read;
	delete attr_PMAX_1_read;
	delete attr_Remote__ServiceMode_read;
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::init_device()
// 
// description : 	will be called at device initialization.
//
//-----------------------------------------------------------------------------
void MenloLFC::init_device()
{
	INFO_STREAM << "MenloLFC::MenloLFC() create device " << device_name << endl;

	// Initialise variables to default values
	//--------------------------------------------
	abortflag = false;
	th_sleep = 1000;
	moxa = false;
	counter = 0;
	
	set_change_event("state", true, false);		//attribute name, implemented, detect, if not setting detect=false have to set abs_change or rel_change for each attribute
	get_device_property();
	dev_add_loc = dev_add + 1;		//define local address as device address + 1
	int found = serialdevice.find(":");
	if(found != string::npos)
	{
		//vector<unsigned int> parsed_ip_cpu;
		//int i=0;
		//unsigned char uc_cpu_ip[4];
		string ip_str(serialdevice.substr(0,found));
/*		string_explode(ip_str, ".", &parsed_ip_cpu);
		if(parsed_ip_cpu.size() == 4)
		{
			for(vector<unsigned int>::iterator it=parsed_ip_cpu.begin(); it != parsed_ip_cpu.end(); it++)
				uc_cpu_ip[i++] = (unsigned char)*it;
		}*/
		strcpy(servIP, ip_str.c_str());
		string port_str(serialdevice.substr(found+1));
		std::istringstream tmp(port_str);
		tmp >> servPort;
		moxa = true;
		if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		{
			/*set_state(Tango::FAULT);
			set_status("socket() failed");
			conn_error = true;*/
			ERROR_STREAM << "MenloLFC::MenloLFC(): socket error";
			exit(1);
		}
		memset(&servAddr, 0, sizeof(servAddr));     	/* Zero out structure {} */
		servAddr.sin_family      = AF_INET;             /* Internet address family */
		servAddr.sin_addr.s_addr = inet_addr(servIP);   /* Server IP address */
		servAddr.sin_port        = htons(servPort); 	/* Server port */
		/* Establish the connection to the echo server */
		if (connect(sock, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0)
		{
			stringstream tmp;
			tmp << "Connection failed to " << servIP  << ":" << servPort << ends;
			set_state(Tango::FAULT);
			set_status(tmp.str());
			conn_error = true;
			INFO_STREAM << "MenloLFC::MenloLFC(): " << tmp.str();
		}
		else
		{
			stringstream tmp;
			tmp << "Connected to " << servIP  << ":" << servPort << ends;
			DEBUG_STREAM << "MenloLFC::MenloLFC(): " << tmp.str();
			conn_error = false;
		}
		
		
	}
	else
	{
		moxa = false;
		Tango::Database *db = new Tango::Database();
		Tango::DbData db_data;	
		try {
			db_data.push_back(Tango::DbDatum("timeout"));
			string in_prop("timeout");
	//		devserial->get_property(in_prop, db_data);
			db->get_device_property(serialdevice,db_data);
			db_data[0] >> serial_timeout;
			delete db;
		}
		catch(Tango::DevFailed e) {
			string temp;
			temp = e.errors[0].desc;
			WARN_STREAM << "MenloLFC::MenloLFC() error getting timeout property from serial device: " << temp << endl;
			delete db;		
		}
		if(serial_timeout < 0 || serial_timeout > 2000)
			serial_timeout = 500;
		/* connect to serialdevice */
		devserial = new Tango::DeviceProxy(serialdevice);	
		
		/* check if serialdevice is responding */
		try {
			if (devserial->ping()) 
			{
				devserial->set_timeout_millis(serial_timeout);
				//set_state(Tango::ON);	
				set_status("Connected to serialdevice " + serialdevice);
				INFO_STREAM << "Connected to serialdevice " + serialdevice << endl;
			}	
		} 
		catch(Tango::DevFailed e) {
			set_state(Tango::FAULT);
			Tango::DevState st = Tango::FAULT;
			push_change_event("state", &st);
			set_status("Failed to connect to serialdevice " + serialdevice);
			INFO_STREAM << "Failed to connect to serialdevice " + serialdevice << endl; 
			conn_error = true;			
		}
		
		Tango::DeviceData Din;
		Tango::DevLong flush = 2;  //flush input and output	
		Din << flush;
		try	{
			devserial->command_inout("DevSerFlush",Din);
		}
		catch(Tango::DevFailed e) {
			string temp;
			temp = e.errors[0].desc;
			WARN_STREAM << "MenloLFC::MenloLFC() error flushing serial device: " << temp << endl;
		}
	}

	attr_ISET_read = new Tango::DevLong();
	attr_IACT_read = new Tango::DevLong();
	attr_PACT_read = new Tango::DevLong();
	attr_DIODE_TSET_read = new Tango::DevLong();
	attr_DIODE_TACT_read = new Tango::DevLong();
	attr_MODULE_TACT_read = new Tango::DevLong();
	attr_TCORR_read = new Tango::DevLong();
	attr_TC0_ADJ_ISET_read = new Tango::DevLong[6];
	attr_TC0_ADJ_TSET_read = new Tango::DevLong[6];
	attr_TC0_ADJ_IACT_read = new Tango::DevLong[4];
	attr_TM0_ADJ_TACT_read = new Tango::DevLong[4];
	attr_TM0_ADJ_PACT_read = new Tango::DevLong[4];
	attr_Amplifier_State_read = new Tango::DevBoolean();
	attr_Snapshot_read = new Tango::DevLong[256*256];
	
	attr_Date_read = new Tango::DevLong[3];
	attr_Time_read = new Tango::DevLong[3];
	attr_RTstamp_read = new Tango::DevDouble();
	attr_Tstamp_read = new Tango::DevDouble();
	attr_TstampDiff_read = new Tango::DevDouble();
	attr_TstampStart_read = new Tango::DevDouble();
	attr_Uptime_read = new Tango::DevDouble();
	
	attr_ADJ_PACT_1_read = new Tango::DevLong[4];
	attr_PACT_1_read = new Tango::DevDouble();
	attr_PMIN_1_read = new Tango::DevDouble();
	attr_PMAX_1_read = new Tango::DevDouble();
	attr_Remote__ServiceMode_read = new Tango::DevLong();
	attr_EnableAutoSave_read = &attr_EnableAutoSave_write;

	menlo = new Menlo(dev_add, dev_add_loc, true);
	mutex = new omni_mutex();
	vector<string> names;
	string reg_s;
	data_t rdata;
	rdata.max_subdev = 1;
	names.push_back(S_DEV);
	names.push_back(S_ADDR);
	menlo->NamesCat(reg_s, names);
	DEBUG_STREAM << "MenloLFC::MenloLFC: prepared name " << reg_s << endl;
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() == 0))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device Address, err=" << rdata.msg << endl;
		set_status("error reading Device Address, err=" + rdata.msg);
	}
	else
	{
		dev_add = rdata.data[0].front().u8;		//update device address with the one read from the device
		dev_add_loc = dev_add + 1;		//define local address as device address + 1
	}
	rdata.data[0].clear();
	abort_sleep(wr_sleep/2000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_ID);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() == 0))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device ID, err=" << rdata.msg << endl;
		set_status("error reading Device ID, err=" + rdata.msg);
		strcpy(dev_id, "");
	}
	else
	{
		strcpy(dev_id, rdata.data[0].front().cstring);
	}
	
	rdata.data[0].clear();
	abort_sleep(wr_sleep/2000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_VER_HW);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 4))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device HW Version, err=" << rdata.msg << endl;
		set_status("error reading Device HW Version, err=" + rdata.msg);
		strcpy(dev_hw_ver, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int major,minor,patchlevel,build;
		deque<data_union>::iterator it = rdata.data[0].begin();
		build = it->u8;
		it++;
		patchlevel = it->u8;
		it++;
		minor = it->u8;
		it++;
		major = it->u8;
		sprintf(dev_hw_ver,"%d,%d,%d,%d", major,minor,patchlevel,build);
	}
	rdata.data[0].clear();
	abort_sleep(wr_sleep/2000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_VER_FW);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 4))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device FW Version, err=" << rdata.msg << endl;
		set_status("error reading Device FW Version, err=" + rdata.msg);
		strcpy(dev_fw_ver, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int major,minor,patchlevel,build;
		deque<data_union>::iterator it = rdata.data[0].begin();
		build = it->u8;
		it++;
		patchlevel = it->u8;
		it++;
		minor = it->u8;
		it++;
		major = it->u8;
		sprintf(dev_fw_ver,"%d,%d,%d,%d", major,minor,patchlevel,build);
	}
	rdata.data[0].clear();
	abort_sleep(wr_sleep/2000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_VER_DISPFW);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 4))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device FW Version, err=" << rdata.msg << endl;
		set_status("error reading Device FW Version, err=" + rdata.msg);
		strcpy(display_fw_ver, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int major,minor,patchlevel,build;
		deque<data_union>::iterator it = rdata.data[0].begin();
		build = it->u8;
		it++;
		patchlevel = it->u8;
		it++;
		minor = it->u8;
		it++;
		major = it->u8;
		sprintf(display_fw_ver,"%d,%d,%d,%d", major,minor,patchlevel,build);
	}
	rdata.data[0].clear();
	abort_sleep(wr_sleep/2000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_SER);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 3))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device Serial, err=" << rdata.msg << endl;
		set_status("error reading Device Serial, err=" + rdata.msg);
		strcpy(dev_serial, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int year,month,serial;
		deque<data_union>::iterator it = rdata.data[0].begin();
		year = it->u8;
		it++;
		month = it->u8;
		it++;
		serial = it->u16;
		sprintf(dev_serial,"%d,%d,%d", year,month,serial);
	}	

	int timeout = (int)(((7) * (serial_timeout+th_sleep+wr_sleep))+5000);

	struct timeval curr_time;
	gettimeofday(&curr_time,NULL);
	//data_union u8((unsigned char)0);
	deque<data_union> dv;
	data_t c_t1 = {1,1,{dv},Hi_pri,"",(byte)(-1),curr_time,timeout};
	data_t c_t2 = {1,1,{dv},Low_pri,"",(byte)(-1),curr_time,timeout};
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_ISET);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t1));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_IACT);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t1));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_PACT);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t1));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_DIODE_TSET);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t1));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_DIODE_TACT);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t1));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_MODULE_TACT);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t1));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_TCORR);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t1));
	names.clear();
	names.push_back(S_DET4);
	names.push_back(S_PACT_1);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t1));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_TC0_ADJ_ISET);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_TC0_ADJ_TSET);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_TC0_ADJ_IACT);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_TM0_ADJ_TACT);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	names.clear();
	names.push_back(S_MDD0);
	names.push_back(S_TM0_ADJ_PACT);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	names.clear();
	names.push_back(S_DET4);
	names.push_back(S_ADJ_PACT_1);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	names.clear();
	names.push_back(S_DET4);
	names.push_back(S_PMIN_1);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	names.clear();
	names.push_back(S_DET4);
	names.push_back(S_PMAX_1);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	
	names.clear();
	names.push_back(S_REMOTE);
	names.push_back(S_STATE_AMP1);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t1));
	
/*	names.clear();
	names.push_back(S_DEV_FILE);
	names.push_back(S_EDIPBMP);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));*/
	
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_TIME);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_DATE);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_RTSTAMP);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_TSTAMP);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_TSTAMPDIFF);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_TSTAMPSTART);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_UPTIME);
	menlo->NamesCat(reg_s, names);	
	read_data.insert(make_pair(reg_s,c_t2));
	
	names.clear();
	names.push_back(S_REMOTE);
	names.push_back(S_SERVICE);
	menlo->NamesCat(reg_s, names);
	read_data.insert(make_pair(reg_s,c_t2));

	int numHi=0, numLow=0;
	for(data_map::iterator it2 = read_data.begin(); it2 != read_data.end(); it2++)
	{
		if(it2->second.pri == Hi_pri)
			numHi++;
		else if(it2->second.pri == Low_pri)
			numLow++;			
	}
	int completeLoop = numHi*PRIORITY_FACTOR + numLow;	
	//int hiPrioLoop = (int) ceil((double)(completeLoop / PRIORITY_FACTOR));
	int loPrioLoop = completeLoop + 1;
	int loLoopPeriod;
	if(numLow > 0)
		loLoopPeriod = (long) (((double)completeLoop) / numLow);
	if(loLoopPeriod==1)		//TODO: if so few high priority reads invert computation (hiLoopPeriod)
		loLoopPeriod++;
	int hiPrioLoop = (int)ceil((numHi + ((double)(numHi))/(loLoopPeriod-1) + 1));
	timeoutHiPrio = (int)(3*((hiPrioLoop+2) * (serial_timeout+th_sleep+wr_sleep))+5000);
	timeoutLowPrio = (int)(3*3*((loPrioLoop+2) * (serial_timeout+th_sleep+wr_sleep))+5000);	// 3* because of force_scheduler
	DEBUG_STREAM << "MenloLFC::MenloLFC(): hiPrioLoop=" << hiPrioLoop << " loPrioLoop=" << loPrioLoop << " timeoutHiPrio=" << timeoutHiPrio << " timeoutLowPrio=" << timeoutLowPrio << endl;
	//update timeout
	for(data_map::iterator it3 = read_data.begin(); it3 != read_data.end(); it3++)
	{
		if(it3->second.pri == Hi_pri)
			it3->second.timeout = timeoutHiPrio;
		else if(it3->second.pri == Low_pri)
			it3->second.timeout = timeoutLowPrio;			
	}


	readloop = new readthread(this);
	readloop->start();
}


//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::get_device_property()
// 
// description : 	Read the device properties from database.
//
//-----------------------------------------------------------------------------
void MenloLFC::get_device_property()
{
	//	Initialize your default values here (if not done with  POGO).
	//------------------------------------------------------------------

	//	Read device properties from database.(Automatic code generation)
	//------------------------------------------------------------------
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("Serialdevice"));
	dev_prop.push_back(Tango::DbDatum("Serial_timeout"));
	dev_prop.push_back(Tango::DbDatum("Th_sleep"));
	dev_prop.push_back(Tango::DbDatum("Wr_sleep"));
	dev_prop.push_back(Tango::DbDatum("Dev_add"));

	//	Call database and extract values
	//--------------------------------------------
	if (Tango::Util::instance()->_UseDb==true)
		get_db_device()->get_property(dev_prop);
	Tango::DbDatum	def_prop, cl_prop;
	MenloLFCClass	*ds_class =
		(static_cast<MenloLFCClass *>(get_device_class()));
	int	i = -1;

	//	Try to initialize Serialdevice from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  serialdevice;
	else {
		//	Try to initialize Serialdevice from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  serialdevice;
	}
	//	And try to extract Serialdevice value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  serialdevice;

	//	Try to initialize Serial_timeout from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  serial_timeout;
	else {
		//	Try to initialize Serial_timeout from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  serial_timeout;
	}
	//	And try to extract Serial_timeout value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  serial_timeout;

	//	Try to initialize Th_sleep from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  th_sleep;
	else {
		//	Try to initialize Th_sleep from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  th_sleep;
	}
	//	And try to extract Th_sleep value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  th_sleep;

	//	Try to initialize Wr_sleep from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  wr_sleep;
	else {
		//	Try to initialize Wr_sleep from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  wr_sleep;
	}
	//	And try to extract Wr_sleep value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  wr_sleep;

	//	Try to initialize Dev_add from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  dev_add;
	else {
		//	Try to initialize Dev_add from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  dev_add;
	}
	//	And try to extract Dev_add value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  dev_add;



	//	End of Automatic code generation
	//------------------------------------------------------------------

}
//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::always_executed_hook()
// 
// description : 	method always executed before any command is executed
//
//-----------------------------------------------------------------------------
void MenloLFC::always_executed_hook()
{

}
//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_attr_hardware
// 
// description : 	Hardware acquisition for attributes.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_attr_hardware(vector<long> &attr_list)
{
	//DEBUG_STREAM << "MenloLFC::read_attr_hardware(vector<long> &attr_list) entering... "<< endl;
	//	Add your own code here

}
//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_IACT
// 
// description : 	Extract real attribute values for IACT acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_IACT(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_IACT(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_IACT);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_IACT_read, /*factor*/1.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_PACT
// 
// description : 	Extract real attribute values for PACT acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_PACT(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_PACT(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_PACT);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_PACT_read, /*factor*/1.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_DIODE_TSET
// 
// description : 	Extract real attribute values for DIODE_TSET acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_DIODE_TSET(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_DIODE_TSET(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_DIODE_TSET);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_DIODE_TSET_read, /*factor*/1.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_DIODE_TSET
// 
// description : 	Write DIODE_TSET attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_DIODE_TSET(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_DIODE_TSET(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_DIODE_TSET);
	menlo->NamesCat(reg_s, names);	
	write_attr_value<const Tango::DevLong *>(attr, reg_s, &attr_DIODE_TSET_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_DIODE_TACT
// 
// description : 	Extract real attribute values for DIODE_TACT acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_DIODE_TACT(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_DIODE_TACT(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_DIODE_TACT);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_DIODE_TACT_read, /*factor*/1.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_MODULE_TACT
// 
// description : 	Extract real attribute values for MODULE_TACT acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_MODULE_TACT(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_MODULE_TACT(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_MODULE_TACT);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_MODULE_TACT_read, /*factor*/1.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_TCORR
// 
// description : 	Extract real attribute values for TCORR acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_TCORR(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_TCORR(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TCORR);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_TCORR_read, /*factor*/1.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_TCORR
// 
// description : 	Write TCORR attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_TCORR(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_TCORR(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TCORR);
	menlo->NamesCat(reg_s, names);	
	write_attr_value<const Tango::DevLong *>(attr, reg_s, &attr_TCORR_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_ISET
// 
// description : 	Extract real attribute values for ISET acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_ISET(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_ISET(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_ISET);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_ISET_read, /*factor*/1.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_ISET
// 
// description : 	Write ISET attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_ISET(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_ISET(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_ISET);
	menlo->NamesCat(reg_s, names);	
	write_attr_value<const Tango::DevLong *>(attr, reg_s, &attr_ISET_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_TC0_ADJ_ISET
// 
// description : 	Extract real attribute values for TC0_ADJ_ISET acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_TC0_ADJ_ISET(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_TC0_ADJ_ISET(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TC0_ADJ_ISET);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_TC0_ADJ_ISET_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/6);
	apply_attr_value_dacs(attr, reg_s, attr_TC0_ADJ_ISET_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_TC0_ADJ_ISET
// 
// description : 	Write TC0_ADJ_ISET attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_TC0_ADJ_ISET(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_TC0_ADJ_ISET(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TC0_ADJ_ISET);
	menlo->NamesCat(reg_s, names);		
	write_attr_value<const Tango::DevLong *>(attr, reg_s, attr_TC0_ADJ_ISET_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_TC0_ADJ_TSET
// 
// description : 	Extract real attribute values for TC0_ADJ_TSET acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_TC0_ADJ_TSET(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_TC0_ADJ_TSET(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TC0_ADJ_TSET);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_TC0_ADJ_TSET_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/6);
	apply_attr_value_dacs(attr, reg_s, attr_TC0_ADJ_TSET_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_TC0_ADJ_TSET
// 
// description : 	Write TC0_ADJ_TSET attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_TC0_ADJ_TSET(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_TC0_ADJ_TSET(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TC0_ADJ_TSET);
	menlo->NamesCat(reg_s, names);		
	write_attr_value<const Tango::DevLong *>(attr, reg_s, attr_TC0_ADJ_TSET_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_TC0_ADJ_IACT
// 
// description : 	Extract real attribute values for TC0_ADJ_IACT acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_TC0_ADJ_IACT(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_TC0_ADJ_IACT(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TC0_ADJ_IACT);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_TC0_ADJ_IACT_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/4);
	apply_attr_value_adcs(attr, reg_s, attr_TC0_ADJ_IACT_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_TC0_ADJ_IACT
// 
// description : 	Write TC0_ADJ_IACT attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_TC0_ADJ_IACT(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_TC0_ADJ_IACT(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TC0_ADJ_IACT);
	menlo->NamesCat(reg_s, names);		
	write_attr_value<const Tango::DevLong *>(attr, reg_s, attr_TC0_ADJ_IACT_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_TM0_ADJ_TACT
// 
// description : 	Extract real attribute values for TM0_ADJ_TACT acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_TM0_ADJ_TACT(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_TM0_ADJ_TACT(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TM0_ADJ_TACT);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_TM0_ADJ_TACT_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/4);
	apply_attr_value_adcs(attr, reg_s, attr_TM0_ADJ_TACT_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_TM0_ADJ_TACT
// 
// description : 	Write TM0_ADJ_TACT attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_TM0_ADJ_TACT(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_TM0_ADJ_TACT(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TM0_ADJ_TACT);
	menlo->NamesCat(reg_s, names);		
	write_attr_value<const Tango::DevLong *>(attr, reg_s, attr_TM0_ADJ_TACT_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_TM0_ADJ_PACT
// 
// description : 	Extract real attribute values for TM0_ADJ_PACT acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_TM0_ADJ_PACT(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_TM0_ADJ_PACT(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TM0_ADJ_PACT);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_TM0_ADJ_PACT_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/4);
	apply_attr_value_adcs(attr, reg_s, attr_TM0_ADJ_PACT_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_TM0_ADJ_PACT
// 
// description : 	Write TM0_ADJ_PACT attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_TM0_ADJ_PACT(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_TM0_ADJ_PACT(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_MDD0);
	names.push_back(S_TM0_ADJ_PACT);
	menlo->NamesCat(reg_s, names);		
	write_attr_value<const Tango::DevLong *>(attr, reg_s, attr_TM0_ADJ_PACT_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_Amplifier_State
// 
// description : 	Extract real attribute values for Amplifier_State acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_Amplifier_State(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_Amplifier_State(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_REMOTE);
	names.push_back(S_STATE_AMP1);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevBoolean>(attr, reg_s, attr_Amplifier_State_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_Amplifier_State
// 
// description : 	Write Amplifier_State attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_Amplifier_State(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_Amplifier_State(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_REMOTE);
	names.push_back(S_STATE_AMP1);
	menlo->NamesCat(reg_s, names);	
	write_attr_value<const Tango::DevBoolean *>(attr, reg_s, &attr_Amplifier_State_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_Snapshot
// 
// description : 	Extract real attribute values for Snapshot acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_Snapshot(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_Snapshot_State(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV_FILE);
	names.push_back(S_EDIPBMP);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_Snapshot_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/256*256);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_Date
// 
// description : 	Extract real attribute values for Date acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_Date(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_Date(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV);
	names.push_back(S_DATE);
	menlo->NamesCat(reg_s, names);	
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_Date_read, /*factor*/1.0, /*index*/-1, /*subdevice*/0, /*size*/3);
	//apply_attr_value_datetime(attr, S_DATE, attr_Date_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_Date
// 
// description : 	Write Date attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_Date(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_Date(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV);
	names.push_back(S_DATE);
	menlo->NamesCat(reg_s, names);	
	write_attr_value<const Tango::DevLong *>(attr, reg_s, attr_Date_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_Time
// 
// description : 	Extract real attribute values for Time acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_Time(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_Time(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV);
	names.push_back(S_TIME);
	menlo->NamesCat(reg_s, names);	
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_Time_read, /*factor*/1.0, /*index*/-1, /*subdevice*/0, /*size*/3);
	//apply_attr_value_datetime(attr, S_DATE, attr_Time_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_Time
// 
// description : 	Write Time attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_Time(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_Time(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV);
	names.push_back(S_TIME);
	menlo->NamesCat(reg_s, names);	
	write_attr_value<const Tango::DevLong *>(attr, reg_s, attr_Time_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_RTstamp
// 
// description : 	Extract real attribute values for Tstamp acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_RTstamp(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_RTstamp(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV);
	names.push_back(S_RTSTAMP);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_RTstamp_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/6);
	apply_attr_value_tstamp(attr, reg_s, attr_RTstamp_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_Tstamp
// 
// description : 	Extract real attribute values for Tstamp acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_Tstamp(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_Tstamp(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV);
	names.push_back(S_TSTAMP);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_Tstamp_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/6);
	apply_attr_value_tstamp(attr, reg_s, attr_Tstamp_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_TstampDiff
// 
// description : 	Extract real attribute values for TstampDiff acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_TstampDiff(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_TstampDiff(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV);
	names.push_back(S_TSTAMPDIFF);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_TstampDiff_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/6);
	apply_attr_value_tstamp(attr, reg_s, attr_TstampDiff_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_TstampStart
// 
// description : 	Extract real attribute values for TstampStart acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_TstampStart(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_TstampStart(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV);
	names.push_back(S_TSTAMPSTART);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_TstampStart_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/6);
	apply_attr_value_tstamp(attr, reg_s, attr_TstampStart_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_Uptime
// 
// description : 	Extract real attribute values for Uptime acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_Uptime(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_Uptime(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DEV);
	names.push_back(S_UPTIME);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_Uptime_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/6);
	apply_attr_value_tstamp(attr, reg_s, attr_Uptime_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_ADJ_PACT_1
// 
// description : 	Extract real attribute values for ADJ_PACT_1 acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_ADJ_PACT_1(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_ADJ_PACT_1(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DET4);
	names.push_back(S_ADJ_PACT_1);
	menlo->NamesCat(reg_s, names);		
	//apply_attr_value<long>(attr, reg_s, attr_ADJ_PACT_1_read, /*factor*/1, /*index*/0, /*subdevice*/0, /*size*/4);
	apply_attr_value_adcs(attr, reg_s, attr_ADJ_PACT_1_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_ADJ_PACT_1
// 
// description : 	Write ADJ_PACT_1 attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_ADJ_PACT_1(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "MenloLFC::write_ADJ_PACT_1(Tango::WAttribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DET4);
	names.push_back(S_ADJ_PACT_1);
	menlo->NamesCat(reg_s, names);		
	write_attr_value<const Tango::DevLong *>(attr, reg_s, attr_ADJ_PACT_1_write,/*subdevice*/-1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_PACT_1
// 
// description : 	Extract real attribute values for PACT_1 acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_PACT_1(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_PACT_1(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DET4);
	names.push_back(S_PACT_1);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevDouble>(attr, reg_s, attr_PACT_1_read, /*factor*/800.0/300.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_PMIN_1
// 
// description : 	Extract real attribute values for PMIN_1 acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_PMIN_1(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_PMIN_1(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DET4);
	names.push_back(S_PMIN_1);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevDouble>(attr, reg_s, attr_PMIN_1_read, /*factor*/800.0/300.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_PMAX_1
// 
// description : 	Extract real attribute values for PMAX_1 acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_PMAX_1(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_PMAX_1(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;
	names.push_back(S_DET4);
	names.push_back(S_PMAX_1);
	menlo->NamesCat(reg_s, names);		
	apply_attr_value<Tango::DevDouble>(attr, reg_s, attr_PMAX_1_read, /*factor*/800.0/300.0, /*index*/0, /*subdevice*/0, /*size*/1);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_Remote__ServiceMode
//
// description : 	Extract real attribute values for Remote__ServiceMode acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_Remote__ServiceMode(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_Remote__ServiceMode(Tango::Attribute &attr) entering... "<< endl;
	vector<string> names;
	string reg_s;

	names.push_back(S_REMOTE);
	names.push_back(S_SERVICE);
	menlo->NamesCat(reg_s, names);
	apply_attr_value<Tango::DevLong>(attr, reg_s, attr_Remote__ServiceMode_read, /*factor*/1.0, /*index*/0, /*subdevice*/0, /*size*/1);

}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::read_EnableAutoSave
//
// description : 	Extract real attribute values for EnableAutoSave acquisition result.
//
//-----------------------------------------------------------------------------
void MenloLFC::read_EnableAutoSave(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::read_EnableAutoSave(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_EnableAutoSave_read);
}

//+----------------------------------------------------------------------------
//
// method : 		MenloLFC::write_EnableAutoSave
//
// description : 	Write EnableAutoSave attribute values to hardware.
//
//-----------------------------------------------------------------------------
void MenloLFC::write_EnableAutoSave(Tango::WAttribute &attr)
{
	//DEBUG_STREAM << "MenloLFC::write_EnableAutoSave(Tango::WAttribute &attr) entering... "<< endl;
	attr.get_write_value(attr_EnableAutoSave_write);
}

//+------------------------------------------------------------------
/**
 *	method:	MenloLFC::device_address
 *
 *	description:	method to execute "DeviceAddress"
 *	Read Device Address
 *
 * @return	Device Address
 *
 */
//+------------------------------------------------------------------
Tango::DevLong MenloLFC::device_address()
{
	Tango::DevLong	argout ;
	DEBUG_STREAM << "MenloLFC::device_address(): entering... !" << endl;

	//	Add your own code to control device here
	argout = dev_add;
	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	MenloLFC::device_id
 *
 *	description:	method to execute "DeviceID"
 *	Read Device ID
 *
 * @return	Device ID
 *
 */
//+------------------------------------------------------------------
Tango::DevString MenloLFC::device_id()
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevString	argout  = new char[6];
	//strcpy(argout, "dummy");
	DEBUG_STREAM << "MenloLFC::device_id(): entering... !" << endl;
	vector<string> names;
	string reg_s;
	data_t rdata;
	rdata.max_subdev = 1;
	rdata.data[0].clear();
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_ID);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() == 0))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device ID, err=" << rdata.msg << endl;
		set_status("error reading Device ID, err=" + rdata.msg);
		strcpy(dev_id, "");
	}
	else
	{
		strcpy(dev_id, rdata.data[0].front().cstring);
	}

	//	Add your own code to control device here
	argout = CORBA::string_dup(dev_id);
	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	MenloLFC::device_hwver
 *
 *	description:	method to execute "DeviceHWver"
 *	Read Device Hw Ver
 *
 * @return	Device HW Ver
 *
 */
//+------------------------------------------------------------------
Tango::DevString MenloLFC::device_hwver()
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevString	argout  = new char[6];
	//strcpy(argout, "dummy");
	DEBUG_STREAM << "MenloLFC::device_hwver(): entering... !" << endl;
	vector<string> names;
	string reg_s;
	data_t rdata;
	rdata.max_subdev = 1;
	rdata.data[0].clear();
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_VER_HW);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 4))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device HW Version, err=" << rdata.msg << endl;
		set_status("error reading Device HW Version, err=" + rdata.msg);
		strcpy(dev_hw_ver, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int major,minor,patchlevel,build;
		deque<data_union>::iterator it = rdata.data[0].begin();
		build = it->u8;
		it++;
		patchlevel = it->u8;
		it++;
		minor = it->u8;
		it++;
		major = it->u8;
		sprintf(dev_hw_ver,"%d,%d,%d,%d", major,minor,patchlevel,build);
	}
	//	Add your own code to control device here
	argout = CORBA::string_dup(dev_hw_ver);
	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	MenloLFC::device_fwver
 *
 *	description:	method to execute "DeviceFWver"
 *	Read Device Fw Ver
 *
 * @return	Device FW Ver
 *
 */
//+------------------------------------------------------------------
Tango::DevString MenloLFC::device_fwver()
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevString	argout  = new char[6];
	//strcpy(argout, "dummy");
	DEBUG_STREAM << "MenloLFC::device_fwver(): entering... !" << endl;
	vector<string> names;
	string reg_s;
	data_t rdata;
	rdata.max_subdev = 1;
	rdata.data[0].clear();
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_VER_FW);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 4))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device FW Version, err=" << rdata.msg << endl;
		set_status("error reading Device FW Version, err=" + rdata.msg);
		strcpy(dev_fw_ver, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int major,minor,patchlevel,build;
		deque<data_union>::iterator it = rdata.data[0].begin();
		build = it->u8;
		it++;
		patchlevel = it->u8;
		it++;
		minor = it->u8;
		it++;
		major = it->u8;
		sprintf(dev_fw_ver,"%d,%d,%d,%d", major,minor,patchlevel,build);
	}
	//	Add your own code to control device here
	argout = CORBA::string_dup(dev_fw_ver);
	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	MenloLFC::display_fwver
 *
 *	description:	method to execute "DisplayFWver"
 *	Read Device Fw Ver
 *
 * @return	Device FW Ver
 *
 */
//+------------------------------------------------------------------
Tango::DevString MenloLFC::display_fwver()
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevString	argout  = new char[6];
	//strcpy(argout, "dummy");
	DEBUG_STREAM << "MenloLFC::display_fwver(): entering... !" << endl;
	vector<string> names;
	string reg_s;
	data_t rdata;
	rdata.max_subdev = 1;
	rdata.data[0].clear();
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_VER_DISPFW);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 4))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device FW Version, err=" << rdata.msg << endl;
		set_status("error reading Device FW Version, err=" + rdata.msg);
		strcpy(display_fw_ver, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int major,minor,patchlevel,build;
		deque<data_union>::iterator it = rdata.data[0].begin();
		build = it->u8;
		it++;
		patchlevel = it->u8;
		it++;
		minor = it->u8;
		it++;
		major = it->u8;
		sprintf(display_fw_ver,"%d,%d,%d,%d", major,minor,patchlevel,build);
	}
	//	Add your own code to control device here
	argout = CORBA::string_dup(display_fw_ver);
	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	MenloLFC::device_serial
 *
 *	description:	method to execute "DeviceSerial"
 *	Read Device Serial
 *
 * @return	Device Serial
 *
 */
//+------------------------------------------------------------------
Tango::DevString MenloLFC::device_serial()
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevString	argout  = new char[6];
	//strcpy(argout, "dummy");
	DEBUG_STREAM << "MenloLFC::device_serial(): entering... !" << endl;
	vector<string> names;
	string reg_s;
	data_t rdata;
	rdata.max_subdev = 1;
	rdata.data[0].clear();
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_SER);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, reg_s, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 3))
	{
		INFO_STREAM << "MenloLFC::MenloLFC: error reading Device Serial, err=" << rdata.msg << endl;
		set_status("error reading Device Serial, err=" + rdata.msg);
		strcpy(dev_serial, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int year,month,serial;
		deque<data_union>::iterator it = rdata.data[0].begin();
		year = it->u8;
		it++;
		month = it->u8;
		it++;
		serial = it->u16;
		sprintf(dev_serial,"%d,%d,%d", year,month,serial);
	}
	//	Add your own code to control device here
	argout = CORBA::string_dup(dev_serial);
	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	MenloLFC::reset_device
 *
 *	description:	method to execute "EnterServiceCode"
 *	Save configuration
 *
 * @return
 *
 */
//+------------------------------------------------------------------
void MenloLFC::enter_service_code(Tango::DevUShort argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	DEBUG_STREAM << "MenloLFC::enter_service_code(): entering... !" << endl;
	vector<string> names;
	string reg_s;
	data_t rdata;
	int subdev=0;
	rdata.max_subdev = 1;
	rdata.data[0].clear();
	names.clear();
	names.push_back(S_REMOTE);
	names.push_back(S_SERVICE);
	menlo->NamesCat(reg_s, names);
	rdata.data[subdev].push_back((uint16_t)argin);
	DEBUG_STREAM << "MenloLFC::enter_service_code(): 4 !" << endl;
	WriteRead(PROTO_CMD_WRITE, reg_s, -1, rdata);
	DEBUG_STREAM << "MenloLFC::enter_service_code(): 5 !" << endl;
	if(rdata.msg != "")
	{
		stringstream out_stream;
		out_stream << "Error entering service code: " << rdata.msg;
		Tango::Except::throw_exception(
			(const char *) "SERVICE_CODE_ERROR",
			out_stream.str(),
			(const char *) "MenloLFC::enter_service_code()", Tango::ERR);
	}
}

//+------------------------------------------------------------------
/**
 *	method:	MenloLFC::save_config
 *
 *	description:	method to execute "SaveConfig"
 *	Save configuration
 *
 * @return
 *
 */
//+------------------------------------------------------------------
void MenloLFC::save_config()
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	DEBUG_STREAM << "MenloLFC::save_config(): entering... !" << endl;
	vector<string> names;
	string reg_s;
	data_t rdata;
	int subdev=0;
	rdata.max_subdev = 1;
	rdata.data[0].clear();
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_SAVSET);
	menlo->NamesCat(reg_s, names);
	rdata.data[subdev].push_back((unsigned char)1);
	WriteRead(PROTO_CMD_WRITE, reg_s, -1, rdata);
	if(rdata.msg != "")
	{
		stringstream out_stream;
		out_stream << "Error saving settings: " << rdata.msg;
		Tango::Except::throw_exception(
			(const char *) "SAVE_ERROR",
			out_stream.str(),
			(const char *) "MenloLFC::save_config()", Tango::ERR);
	}
}


int MenloLFC::WriteRead(int cmd, string s_reg, int subdevice, data_t &rdata)
{
	mutex->lock();
//	DEBUG_STREAM << "MenloLFC::WriteRead(): entering... !" << endl;
	char outbuf[256],
	inbuf[256];
	stringstream temp;
	//data_map::iterator it;
	//string cmd = it->first;
	bool ok = false;
	int nretries = 0;
	long read_len=0;
	long write_len;
	
	/*it = read_data.find(cmd);
	if(it == read_data.end())
	{
		ERROR_STREAM << "MenloLFC::WriteRead() command not found: " << cmd << endl;
		return -1;
	}*/
	
	//sprintf(outbuf,"%s%s%s",cmd.c_str(),arg.c_str(),"\xa");
	
	
	
	
	if(conn_error)
	{
#if 0
		data_t tmp;
		tmp.data[0].push_back((int)counter);
		write_len = menlo->PrepareMsg(PROTO_CMD_ECHO, "", /*subdevice*/-1, tmp, outbuf);
		if(write_len < 0)
		{
			counter++;
			mutex->unlock();
			return -1;
		}
		dump_msg("MenloLFC::WriteRead write echo:", write_len, outbuf);
		try {
			if(!moxa)
				write_serial(outbuf,write_len);
			else
				write_socket(outbuf,write_len);	
			abort_sleep(wr_sleep/1000);  //abort_sleep(sec)	
			//read_serial(inbuf);			
		}
		catch(Tango::DevFailed &e) {
			TangoSys_MemStream out_stream;
			out_stream << "Error writing cmd = " << PROTO_CMD_ECHO << ", err = " << e.errors[0].desc << ends;
			WARN_STREAM << "MenloLFC::WriteRead(): " << out_stream.str() << endl;
			rdata.msg = out_stream.str();
			counter++;
			mutex->unlock();
			return -1;							
		}
		tmp.data[0].clear();
		memset(inbuf,0,256);	
		try {
			if(!moxa)
				read_serial(inbuf,read_len);
			else
				read_socket(inbuf,read_len);		
		}
		catch(Tango::DevFailed &e) {
			if(get_state() != Tango::FAULT)
			{
				Tango::DevState st = Tango::FAULT;
				set_state(st);	
				push_change_event("state", &st);
			}
			set_status("Errors from device" + serialdevice);	
			TangoSys_MemStream out_stream;
			out_stream << "Error reading cmd = " << PROTO_CMD_ECHO << ", err = " << e.errors[0].desc << ends;	
			WARN_STREAM << "MenloLFC::WriteRead(): " << out_stream.str() << endl;
			tmp.msg = out_stream.str();
			counter++;
			mutex->unlock();
			return -1;
		}
		dump_msg("MenloLFC::WriteRead read echo:", read_len, inbuf);
		menlo->ParseResponse(inbuf,read_len, tmp, 0);
		if(tmp.data[0].size() == 0)
		{
			INFO_STREAM << "MenloLFC::WriteRead: error parsing echo response, size==0" << endl;
			counter++;
			if(get_state() != Tango::FAULT)
			{
				Tango::DevState st = Tango::FAULT;
				set_state(st);	
				push_change_event("state", &st);
			}
			set_status("Errors from device" + serialdevice);	
			mutex->unlock();
			return -1;	
		}
		else
		{
			if(tmp.data[0][0].s32 == counter++)	//subdevice=0, index=0
			{
				if(get_state() != Tango::RUNNING)
				{
					Tango::DevState st = Tango::RUNNING;
					set_state(st);	
					push_change_event("state", &st);
				}
				set_status("Reconnected to device" + serialdevice);	
				conn_error = false;
				reconnect();
			}
			else
			{
				if(get_state() != Tango::FAULT)
				{
					Tango::DevState st = Tango::FAULT;
					set_state(st);	
					push_change_event("state", &st);
				}
				set_status("Errors from serialdevice" + serialdevice);	
				mutex->unlock();
				return -1;	
			}
		}
#else
	
#endif
	}

	
	
	write_len = menlo->PrepareMsg(cmd, s_reg, subdevice, rdata, outbuf);
	if(write_len < 0)
	{
		mutex->unlock();
		return -1;
	}
	//DEBUG_STREAM << "MenloLFC::WriteRead now1 reg_type=" << hex << (int)rdata.reg_type << dec;
	dump_msg("MenloLFC::WriteRead write cmd:", write_len, outbuf);
	try {
		if(!moxa)
			write_serial(outbuf,write_len);
		else
			write_socket(outbuf,write_len);	
		abort_sleep(wr_sleep/1000);  //abort_sleep(sec)	
		//read_serial(inbuf);			
	}
	catch(Tango::DevFailed &e) {
		TangoSys_MemStream out_stream;
		out_stream << "Error writing cmd = " << cmd << ", err = " << e.errors[0].desc << ends;
		WARN_STREAM << "MenloLFC::WriteRead(): " << out_stream.str() << endl;
		rdata.msg = out_stream.str();
		mutex->unlock();
		return -1;							
	}
	DEBUG_STREAM << "MenloLFC::WriteRead now reg_type=" << hex << (int)rdata.reg_type << dec;
	memset(inbuf,0,256);	
	try {
		if(!moxa)
			read_serial(inbuf,read_len);
		else
			read_socket(inbuf,read_len);
		dump_msg("MenloLFC::WriteRead read resp:", read_len, inbuf);
		menlo->ParseResponse(inbuf,read_len, rdata, subdevice);			
	}
	catch(Tango::DevFailed &e) {
		TangoSys_MemStream out_stream;
		out_stream << "Error reading cmd = " << cmd << ", err = " << e.errors[0].desc << ends;	
		WARN_STREAM << "MenloLFC::WriteRead(): " << out_stream.str() << endl;
		rdata.msg = out_stream.str();
		read_len = 0;
	}
			
	mutex->unlock();
	return 0;
}



//+------------------------------------------------------------------
//
//	method:			MenloLFC::write_serial
//
//	description:	method to execute "write_serial"
//					Write a string to serial interface
//
//-------------------------------------------------------------------
void MenloLFC::write_serial(char *write_str, long size)
//void Chiller::write_serial(string write_str)
{
	//DEBUG_STREAM << "MenloLFC::write_serial(): entering... =" << write_str << endl;
	//	Add your own code to control device here
	Tango::DeviceData Din,Din2;
	int i;
	Tango::DevLong flush = 2;
	Tango::DevVarCharArray *input = new Tango::DevVarCharArray();
	input->length(size);
//	input->length(write_str.length());
	stringstream temp;
	
//	for (int i=0; i < write_str.length(); i++)
	for (int i=0; i < size; i++)
	{
		//(*input)[i] = (unsigned char) write_str.at(i);
		(*input)[i] = write_str[i];
		temp << showbase << hex << (int)(*input)[i] << " ";
	}
	//DEBUG_STREAM << "MenloLFC::write_serial(): writing=" << write_str << "=" << temp.str() << endl;	
	DEBUG_STREAM << "MenloLFC::write_serial(): writing=" << temp.str() << endl;

	try {
		Din << input;
		devserial->command_inout("DevSerWriteChar",Din);
/*		Din2 << flush;
		devserial->command_inout("DevSerFlush",Din2);*/
	}
	catch(Tango::DevFailed e) {	
			TangoSys_MemStream out_stream;
			conn_error = true;
			out_stream << "Write command to SerialDeviceServer failed" << ends;
			if(get_state() != Tango::FAULT)
			{
				set_state(Tango::FAULT);
				Tango::DevState st = Tango::FAULT;
				push_change_event("state", &st);
			}
			set_status("Errors from serialdevice " + serialdevice);		
			Tango::Except::re_throw_exception(e,
				(const char *) "Error writing serial",
				out_stream.str(),
				(const char *) "MenloLFC::write_serial()", Tango::ERR);
	}
//	set_state(Tango::ON);
//	set_status("Connected to serialdevice " + serialdevice);
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::write_socket
//
//	description:	method to execute "write_serial"
//					Write a string to serial interface
//
//-------------------------------------------------------------------
void MenloLFC::write_socket(char *write_str, long size)
{
	int ret;
	if(conn_error)
	{
		close (sock);
		if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		{
			/*set_state(Tango::FAULT);
			set_status("socket() failed");
			conn_error = true;*/
			ERROR_STREAM << "MenloLFC::write_socket(): socket error";
			stringstream tmp;
			tmp << "Creation of socket failed" << ends;
			set_state(Tango::FAULT);
			set_status(tmp.str());
			conn_error = true;
			INFO_STREAM << "MenloLFC::write_socket(): " << tmp.str();
			Tango::Except::throw_exception(
				(const char *) "Error writing socket",
				tmp.str(),
				(const char *) "MenloLFC::write_socket()", Tango::ERR);
		}
		memset(&servAddr, 0, sizeof(servAddr));     	/* Zero out structure {} */
		servAddr.sin_family      = AF_INET;             /* Internet address family */
		servAddr.sin_addr.s_addr = inet_addr(servIP);   /* Server IP address */
		servAddr.sin_port        = htons(servPort); 	/* Server port */
		if (connect(sock, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0)
		{
			stringstream tmp;
			tmp << "Connection failed to " << servIP << ":" << servPort << ends;
			set_state(Tango::FAULT);
			set_status(tmp.str());
			conn_error = true;
			INFO_STREAM << "MenloLFC::write_socket(): " << tmp.str();
			Tango::Except::throw_exception(
				(const char *) "Error writing socket",
				tmp.str(),
				(const char *) "MenloLFC::write_socket()", Tango::ERR);
			//return;
		}
	}
	ret = send(sock, write_str, size, 0);
	if(ret == -1)
	{
			stringstream tmp;
			tmp << "Write failed to " << servIP << ":" << servPort << " errno=" << errno << "->" << strerror(errno)<< ends;
			set_state(Tango::FAULT);
			set_status(tmp.str());
			conn_error = true;
			INFO_STREAM << "MenloLFC::write_socket(): " << tmp.str();
			Tango::Except::throw_exception(
				(const char *) "Error writing socket",
				tmp.str(),
				(const char *) "MenloLFC::write_socket()", Tango::ERR);
			//return;	
	}
	else if(get_state() != Tango::ON)
	{
		stringstream tmp;
		tmp << "Connected to " << servIP << ":" << servPort << ends;
		set_state(Tango::ON);
		set_status(tmp.str());
		conn_error = false;
	}
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::read_serial
//
//	description:	method to execute "read_serial"
//					Read from serial interface 
//
//-------------------------------------------------------------------
void MenloLFC::read_serial(char *read_str, long &read_len)
{
	//DEBUG_STREAM << "MenloLFC::write_serial(): entering... !" << endl;
	//	Add your own code to control device here
	Tango::DevVarLongStringArray strin;
	string input, strout, strout_filt;
	Tango::DeviceData Din,Dout;
	const Tango::DevVarCharArray *out;
	//long attr_read_param = SL_RAW;
	long attr_read_param;
	//attr_read_param = SAALEN << 8;
	//attr_read_param = 1 << 8;   //provo a leggere solo un carattere
	//attr_read_param |= SL_NCHAR;
	attr_read_param = SL_LINE;
	
	//attr_read_param = 2 << 8;   //provo a leggere per 2 tentativi
	//attr_read_param |= SL_RETRY;
	long temp_len;
	int i;

	try {	
		Din << attr_read_param;
		Dout = devserial->command_inout("DevSerReadChar",Din);
	}
	catch(Tango::DevFailed e) {
		TangoSys_MemStream out_stream;
		conn_error = true;
		out_stream << "DevSerReadNChar command to SerialDeviceServer failed" << ends;
		//set_state(Tango::FAULT);
		//set_status("Errors from serialdevice " + serialdevice);
		WARN_STREAM << "MenloLFC::read_serial(): " <<  out_stream.str() << endl;			
		Tango::Except::re_throw_exception(e,
			(const char *) "Error reading serial",
			out_stream.str(),
			(const char *) "MenloLFC::read_serial()", Tango::ERR);
	}
	if(!Dout.is_empty())
		Dout >> out;
	else
	{
		//set_state(Tango::FAULT);
		//set_status("Errors from serialdevice " + serialdevice);			
		Tango::Except::throw_exception(
			(const char *) "Serialline timeout",
			(const char *) "No data from serialline",
			(const char *) "MenloLFC::read_serial()", Tango::ERR);		
	}

	memset(read_str,0,256);
	temp_len = out->length();
	if(temp_len > 256)	
	{
		WARN_STREAM << "MenloLFC::read_serial(): read buffer overflow, " << 
			temp_len << " > " <<  256 << endl;
		temp_len = 256;
	}
	read_len = temp_len;
	stringstream temp;
	for (int i = 0; i < temp_len; i++)
	{
		read_str[i] = (*out)[i];
		temp << showbase << hex << (int)(*out)[i] << " ";
	}
	DEBUG_STREAM << "MenloLFC::read_serial(): read=" << temp.str() << endl;
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::read_socket
//
//	description:	method to execute "read_serial"
//					Read from serial interface 
//
//-------------------------------------------------------------------
void MenloLFC::read_socket(char *read_str, long &read_len)
{
	int ret;
//	char tmp_read[64];
	int i=0;
	int len=0;
	DEBUG_STREAM << "MenloLFC::read_socket(): entering conn_error=" << (int)conn_error << endl;
	if(conn_error)
	{
		close (sock);
		if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		{
			/*set_state(Tango::FAULT);
			set_status("socket() failed");
			conn_error = true;*/
			ERROR_STREAM << "MenloLFC::read_socket(): socket error";
			stringstream tmp;
			tmp << "Creation of socket failed" << ends;
			set_state(Tango::FAULT);
			set_status(tmp.str());
			conn_error = true;
			INFO_STREAM << "MenloLFC::read_socket(): " << tmp.str();
			Tango::Except::throw_exception(
				(const char *) "Error reading socket",
				tmp.str(),
				(const char *) "MenloLFC::read_socket()", Tango::ERR);
		}
		memset(&servAddr, 0, sizeof(servAddr));     	/* Zero out structure {} */
		servAddr.sin_family      = AF_INET;             /* Internet address family */
		servAddr.sin_addr.s_addr = inet_addr(servIP);   /* Server IP address */
		servAddr.sin_port        = htons(servPort); 	/* Server port */
		if (connect(sock, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0)
		{
			stringstream tmp;
			tmp << "Connection failed to " << servIP << ":" << servPort << ends;
			set_state(Tango::FAULT);
			set_status(tmp.str());
			conn_error = true;
			INFO_STREAM << "MenloLFC::read_socket(): " << tmp.str();
			Tango::Except::throw_exception(
				(const char *) "Error reading socket",
				tmp.str(),
				(const char *) "MenloLFC::read_socket()", Tango::ERR);
			//return;
		}
	}
	do
	{
		DEBUG_STREAM << "MenloLFC::read_socket(): before  recv 256-len=" << 256-len << endl;
		//ret = recv(sock, &read_str[len], 256-len, 0);
		ret = recv(sock, &read_str[len], 256-len, MSG_DONTWAIT);
		DEBUG_STREAM << "MenloLFC::read_socket(): after  recv ret=" << ret << endl;
		if(ret == -1)
		{
				/*stringstream tmp;
				tmp << "Read failed from " << servIP << ":" << servPort << " errno=" << errno << "->" << strerror(errno)<< ends;
				set_state(Tango::FAULT);
				set_status(tmp.str());
				conn_error = true;
				INFO_STREAM << "MenloLFC::read_socket(): " << tmp.str();
				Tango::Except::throw_exception(
					(const char *) "Error reading socket",
					tmp.str(),
					(const char *) "MenloLFC::read_socket()", Tango::ERR);*/
				abort_sleep(2*wr_sleep/1000);
				//return;	
		}
		else if(get_state() != Tango::ON)
		{
			stringstream tmp;
			tmp << "Connected to " << servIP << ":" << servPort << ends;
			set_state(Tango::ON);
			set_status(tmp.str());
			conn_error = false;
			len += ret;
		}
		else
		{
			len += ret;
		}
		i++;
	}
	while(len == 0 && i < 8);
	if(len == 0)
	{
		stringstream tmp;
		tmp << "Read failed from " << servIP << ":" << servPort << " errno=" << errno << "->" << strerror(errno)<< ends;
		set_state(Tango::FAULT);
		set_status(tmp.str());
		conn_error = true;
		INFO_STREAM << "MenloLFC::read_socket(): " << tmp.str();
		Tango::Except::throw_exception(
			(const char *) "Error reading socket",
			tmp.str(),
			(const char *) "MenloLFC::read_socket()", Tango::ERR);
	}
	read_len = len;
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::apply_attr_value
//
//	description:	
//
//-------------------------------------------------------------------
template <typename Attr_type> void MenloLFC::apply_attr_value(Tango::Attribute &attr, string reg, Attr_type/*Tango::DevDouble*/ *attr_read, double factor, int index, int subdevice, int size)
{
	data_map::iterator it = read_data.find(reg);
	if (it == read_data.end()) 
	{
		INFO_STREAM << "MenloLFC::apply_attr_value: error reading: reg=0x" << hex <<  reg << " not found!" << endl;
		Tango::Except::throw_exception(
			"Error not found reg=" + reg,
			"Error not found reg=" + reg,
			(const char *) "MenloLFC::apply_attr_value", Tango::ERR);			
	}
	bool timeout = false;
	vector<int> types;
	int num_types;
	byte reg_type = it->second.reg_type;
	//DEBUG_STREAM << "MenloLFC::apply_attr_value: found it->first=" << it->first<<" reg_type=0x" << hex << (int)reg_type << dec << endl;
	cmd_type_map::iterator it2;
	int id_type;
	if (check_timeout(it->second.time, it->second.timeout) < 0) {
		if (it->second.msg != "") {
			INFO_STREAM << "MenloLFC::apply_attr_value: error reading: reg=" <<  it->first << " reg_type=0x" << hex << (int)reg_type << " err=" << it->second.msg << endl;
			Tango::Except::throw_exception(
				"Error reading reg=" + it->first,
				(const char *) it->second.msg.c_str(),
				(const char *) "MenloLFC::apply_attr_value", Tango::ERR);			
		}
		else {
			DEBUG_STREAM << "MenloLFC::apply_attr_value: reg=" <<  it->first << " reg_type=0x" << hex << (int)reg_type << " data is not refreshed" << dec << endl;
			timeout = true;					
		}
	}

	it2 = menlo->cmd_type.find(reg_type);
	if(it2 == menlo->cmd_type.end())
	{
		stringstream tmp;
		tmp << "Register " <<  it->first << " reg_type=0x" << hex << (int)reg_type  << dec << " not found!!" << ends;
		Tango::Except::throw_exception(
			"Error reading " + it->first,
			tmp.str(),
			(const char *) "MenloLFC::apply_attr_value", Tango::ERR);
	}
	types = it2->second;
	num_types = types.size();
	if(index == -1)
	{
		if(it->second.data[subdevice].size() < num_types)
		{
			attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_INVALID, num_types, 0, false);
			INFO_STREAM << "MenloLFC::apply_attr_value: INVALID: index == -1 and it->second.data[subdevice].size()=" << it->second.data[subdevice].size() << " < num_types=" << num_types << endl;
			return;
		}
	}
	else if(it->second.data[subdevice].size() < index)
	{
		attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_INVALID, num_types, 0, false);
		INFO_STREAM << "MenloLFC::apply_attr_value: INVALID: it->second.data[subdevice].size()=" << it->second.data[subdevice].size() << " < index=" << index << endl;
		return;
	}
	int attr_size = (size <= num_types) ? size : num_types;
	//DEBUG_STREAM << "MenloLFC::apply_attr_value: attr_size=" << attr_size;
	if(index == -1)		//set size values
	{
		for(int i=0; i<attr_size; i++)
		{
			id_type = types[i];
			switch(id_type)
			{
				case U8:
					attr_read[i] = (Attr_type)(it->second.data[subdevice][i].u8*factor);
				break;
				case U16:
					attr_read[i] = (Attr_type)(it->second.data[subdevice][i].u16*factor);
				break;
				case U32:
					attr_read[i] = (Attr_type)(it->second.data[subdevice][i].u32*factor);
				break;
				case S16:
					attr_read[i] = (Attr_type)(it->second.data[subdevice][i].s16*factor);
				break;
				case S32:
					attr_read[i] = (Attr_type)(it->second.data[subdevice][i].s32*factor);
				break;
			}
		}
	}
	else
	{
		id_type = types[index];
		//DEBUG_STREAM << "MenloLFC::apply_attr_value: 222 id_type=" << id_type;
		switch(id_type)
		{
			case U8:
				*attr_read = (Attr_type)(it->second.data[subdevice][index].u8*factor);
			break;
			case U16:
				*attr_read = (Attr_type)(it->second.data[subdevice][index].u16*factor);
				//DEBUG_STREAM << "MenloLFC::apply_attr_value: 222 read U16=" << attr_read;
			break;
			case U32:
				*attr_read = (Attr_type)(it->second.data[subdevice][index].u32*factor);
			break;
			case S16:
				*attr_read = (Attr_type)(it->second.data[subdevice][index].s16*factor);
				//DEBUG_STREAM << "MenloLFC::apply_attr_value: 222 read S16=" << attr_read;
			break;
			case S32:
				*attr_read = (Attr_type)(it->second.data[subdevice][index].s32*factor);
			break;
		}
	}
	
	if(index == -1)
	{
		if(!timeout)
			attr.set_value(attr_read,attr_size);
		else
			attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_WARNING, attr_size, 0, false);
	}
	else
	{
		if(!timeout)
			attr.set_value(attr_read,1);
		else
			attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_WARNING, 1, 0, false);	
	}
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::apply_attr_value_state
//
//	description:	
//
//-------------------------------------------------------------------
void MenloLFC::apply_attr_value_state(Tango::Attribute &attr, string reg, Tango::DevLong *attr_read, int index)
{
	data_map::iterator it = read_data.find(reg);
	int subdevice=0;
	if (it == read_data.end()) 
	{
		INFO_STREAM << "MenloLFC::apply_attr_value_state: error reading: reg=0x" << hex <<  reg << " not found!" << endl;
		Tango::Except::throw_exception(
			"Error not found reg=" + reg,
			"Error not found reg=" + reg,
			(const char *) "MenloLFC::apply_attr_value_state", Tango::ERR);			
	}
	bool timeout = false;
	if (check_timeout(it->second.time, it->second.timeout) < 0) {
		if (it->second.msg != "") {
			INFO_STREAM << "MenloLFC::apply_attr_value_state: error reading: reg=0x" << hex <<  it->first << " err=" << it->second.msg << endl;
			Tango::Except::throw_exception(
				"Error reading reg=" + it->first,
				(const char *) it->second.msg.c_str(),
				(const char *) "MenloLFC::apply_attr_value_state", Tango::ERR);			
		}
		else {
			DEBUG_STREAM << "MenloLFC::apply_attr_value_state: reg=0x" << hex <<  it->first << " data is not refreshed" << dec << endl;
			timeout = true;					
		}
	}

	if(it->second.data[subdevice].size() < index)
	{
		attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_INVALID, 1, 0, false);
		INFO_STREAM << "MenloLFC::apply_attr_value_state: INVALID: it->second.data[subdevice].size()=" << it->second.data[subdevice].size() << " < index=" << index << endl;
		return;
	}

	switch(it->second.data[subdevice][index].u16)
	{
		case 0:			//OFF
			*attr_read = (Tango::DevLong)0;
			break;
		case 0x10:		//ON
			*attr_read = (Tango::DevLong)1;
			break;
		case 0x1:		//off-to-on transition step 1	
			*attr_read = (Tango::DevLong)2;
			break;
		case 0x2:		//off-to-on transition step 2
			*attr_read = (Tango::DevLong)3;
			break;
		case 0x11:		//on-to-off transition step 1
			*attr_read = (Tango::DevLong)4;
			break;
		case 0x12:		//on-to-off transition step 2
			*attr_read = (Tango::DevLong)5;
			break;
		case 0x80:		//initialization
			*attr_read = (Tango::DevLong)6;
			break;
		case 0x08:		//fault
			*attr_read = (Tango::DevLong)-1;
			break;
	}

	

	if(!timeout)
		attr.set_value(attr_read,1);
	else
		attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_WARNING, 1, 0, false);	
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::apply_attr_value_adcs
//
//	description:	
//
//-------------------------------------------------------------------
void MenloLFC::apply_attr_value_dacs(Tango::Attribute &attr, string reg, Tango::DevLong *attr_read)
{
	data_map::iterator it = read_data.find(reg);
	bool timeout = false;
	int subdevice = 0;
	if (check_timeout(it->second.time, it->second.timeout) < 0) {
		if (it->second.msg != "") {
			Tango::Except::throw_exception(
				"Error reading " + it->first,
				(const char *) it->second.msg.c_str(),
				(const char *) "MenloLFC::apply_attr_value_dacs", Tango::ERR);			
		}
		else {
			DEBUG_STREAM << "MenloLFC::apply_attr_value_dacs: " << it->first << " data is not refreshed" << endl;
			timeout = true;					
		}
	}
	if(it->second.data[subdevice].size() < 6)
	{
		attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_INVALID, 6, 0, false);
		return;
	}

	attr_read[0] = ((Tango::DevLong)it->second.data[subdevice][0].s32);
	attr_read[1] = ((Tango::DevLong)it->second.data[subdevice][1].s32);
	attr_read[2] = ((Tango::DevLong)it->second.data[subdevice][2].s32);
	attr_read[3] = ((Tango::DevLong)it->second.data[subdevice][3].s32);
	attr_read[4] = ((Tango::DevLong)it->second.data[subdevice][4].u16);
	attr_read[5] = ((Tango::DevLong)it->second.data[subdevice][5].u16);


	if(!timeout)
		attr.set_value(attr_read,6);
	else
		attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_WARNING, 6, 0, false);
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::apply_attr_value_adcs
//
//	description:	
//
//-------------------------------------------------------------------
void MenloLFC::apply_attr_value_adcs(Tango::Attribute &attr, string reg, Tango::DevLong *attr_read)
{
	data_map::iterator it = read_data.find(reg);
	bool timeout = false;
	int subdevice = 0;
	if(it == read_data.end())
	{
		Tango::Except::throw_exception(
			"Error reading " + reg,
			reg + " not present in polling thread",
			(const char *) "MenloLFC::apply_attr_value_adcs", Tango::ERR);		
	}
	if (check_timeout(it->second.time, it->second.timeout) < 0) {
		if (it->second.msg != "") {
			Tango::Except::throw_exception(
				"Error reading " + it->first,
				(const char *) it->second.msg.c_str(),
				(const char *) "MenloLFC::apply_attr_value_adcs", Tango::ERR);			
		}
		else {
			DEBUG_STREAM << "MenloLFC::apply_attr_value_adcs: " << it->first << " data is not refreshed" << endl;
			timeout = true;					
		}
	}
	if(it->second.data[subdevice].size() < 4)
	{
		attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_INVALID, 4, 0, false);
		return;
	}

	attr_read[0] = ((Tango::DevLong)it->second.data[subdevice][0].s32);
	attr_read[1] = ((Tango::DevLong)it->second.data[subdevice][1].s32);
	attr_read[2] = ((Tango::DevLong)it->second.data[subdevice][2].s32);
	attr_read[3] = ((Tango::DevLong)it->second.data[subdevice][3].s32);

	if(!timeout)
		attr.set_value(attr_read,4);
	else
		attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_WARNING, 4, 0, false);
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::apply_attr_value_tstamp
//
//	description:	
//
//-------------------------------------------------------------------
void MenloLFC::apply_attr_value_tstamp(Tango::Attribute &attr, string reg, Tango::DevDouble *attr_read)
{
	data_map::iterator it = read_data.find(reg);
	bool timeout = false;
	int subdevice = 0;
	int tmp_s, tmp_ms;
	if(it == read_data.end())
	{
		Tango::Except::throw_exception(
			"Error reading " + reg,
			reg + " not present in polling thread",
			(const char *) "MenloLFC::apply_attr_value_tstamp", Tango::ERR);		
	}
	if (check_timeout(it->second.time, it->second.timeout) < 0) {
		if (it->second.msg != "") {
			Tango::Except::throw_exception(
				"Error reading " + it->first,
				(const char *) it->second.msg.c_str(),
				(const char *) "MenloLFC::apply_attr_value_tstamp", Tango::ERR);			
		}
		else {
			DEBUG_STREAM << "MenloLFC::apply_attr_value_tstamp: " << it->first << " data is not refreshed" << endl;
			timeout = true;					
		}
	}
	if(it->second.data[subdevice].size() < 2)
	{
		attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_INVALID, 1, 0, false);
		return;
	}

	tmp_s = ((Tango::DevLong)it->second.data[subdevice][0].s32);
	tmp_ms = ((Tango::DevLong)it->second.data[subdevice][1].s16);

	*attr_read = (double)(tmp_s + ((double)(tmp_ms))/1000.0);

	if(!timeout)
		attr.set_value(attr_read);
	else
		attr.set_value_date_quality(attr_read,0/*gettime()*/,Tango::ATTR_WARNING, 1, 0, false);
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::write_attr_value
//
//	description:	
//
//-------------------------------------------------------------------
template <typename Attr_type> void MenloLFC::write_attr_value(Tango::WAttribute &attr, string reg, Attr_type/*Tango::DevLong*/ attr_write, int subdevice)
{
	DEBUG_STREAM << "MenloLFC::write_attr_value(Tango::WAttribute &attr) entering... "<< endl;
	vector<int> types;
	int num_types;
	cmd_type_map::iterator it;
	int id_type;
	attr.get_write_value(attr_write);
	data_t rdata;
	int subdev=0;
	int i=0;
	int ret;
	byte reg_type;
	ret = menlo->FindRegType(reg, reg_type);
	if(ret == -1)
	{
		stringstream tmp;
		tmp << "reg type for register " << reg  << " not found!!" << ends;
		INFO_STREAM << "MenloLFC::write_attr_value(): " << tmp.str() << endl; 
		Tango::Except::throw_exception(
			(const char *) "Error",
			tmp.str(),
			(const char *) "MenloLFC::write_attr_value()", Tango::ERR);	
	}
	it = menlo->cmd_type.find(reg_type);
	if(it == menlo->cmd_type.end())
	{
		stringstream tmp;
		tmp << "register " << reg << ", reg_type 0x" << hex << (int)reg_type  << dec << " not found!!" << ends;
		INFO_STREAM << "MenloLFC::write_attr_value(): " << tmp.str() << endl; 
		Tango::Except::throw_exception(
			(const char *) "Error",
			tmp.str(),
			(const char *) "MenloLFC::write_attr_value()", Tango::ERR);
	}
	types = it->second;
	num_types = types.size();
	
	if(attr.get_write_value_length() != num_types)
	{
		stringstream tmp;
		tmp << "Expeting " << num_types << " values" << ends;
		INFO_STREAM << "MenloLFC::write_attr_value(): " << tmp.str() << endl; 
		Tango::Except::throw_exception(
			(const char *) "Error",
			tmp.str(),
			(const char *) "MenloLFC::write_attr_value()", Tango::ERR);	
	}
	if(subdevice != -1)
	{
		subdev = subdevice;
		rdata.max_subdev = 2;	//to enable writing of subdev (if > 1)
		rdata.num_subdev = -1;
	}
	else
	{
		rdata.max_subdev = 1;
		rdata.num_subdev = 1;
	}
	for(int y=0; y<num_types; y++)
	{
		id_type = types[y];
		DEBUG_STREAM << "MenloLFC::write_attr_value(): writing id_type=" << id_type << endl; 
		switch(id_type)
		{
			case U8:
			{
				rdata.data[subdev].push_back((unsigned char)attr_write[i++]);
			}
			break;
					
			case U16:
			{
				rdata.data[subdev].push_back((unsigned short)attr_write[i++]);

			}
			break;
					
			case U32:
			{
				rdata.data[subdev].push_back((unsigned int)attr_write[i++]);
		
			}
			break;
			
			case S16:
			{
				rdata.data[subdev].push_back((short)attr_write[i++]);

			}
			break;
					
			case S32:
			{
				rdata.data[subdev].push_back((int)attr_write[i++]);
			}
			break;
		}
		
	}
	
	
	WriteRead(PROTO_CMD_WRITE, reg, subdevice, rdata);
	if(rdata.msg != "")
	{
		INFO_STREAM << "MenloLFC::write_attr_value: error writing, err=" << rdata.msg << endl;
		Tango::Except::throw_exception(
				"Error writing",
				(const char *) rdata.msg.c_str(),
				(const char *) "MenloLFC::write_attr_value", Tango::ERR);
	}
	readloop->force_scheduler(reg);
	if(*attr_EnableAutoSave_read)
	{
		save_config();
	}
}

//+------------------------------------------------------------------
//
//	method:			MenloLFC::check_timeout
//
//	description:	method to execute "check_timeout"
//					Return -1 if timeout [msec.] exceeded
//
//-------------------------------------------------------------------
int MenloLFC::check_timeout(timeval time, double timeout)
{
	struct timeval timenow;
	gettimeofday(&timenow,NULL);
	
	if ((timenow.tv_sec -	time.tv_sec) > (timeout/1000)) {
		DEBUG_STREAM << "MenloLFC::check_timeout(): TIMEOUT! sec diff=" << (timenow.tv_sec - time.tv_sec) << " > timeout=" << (timeout/1000) << endl;
		return -1;
	}	
	if (((timenow.tv_sec -	time.tv_sec)*1000 - 
			(int)((double)(timenow.tv_usec -	time.tv_usec)/1000)) > timeout) {		
		return -1;
	}
	return 0;
	
}	/* end check_timeout() */	


//+------------------------------------------------------------------
//
//	method:			MenloLFC::abort_sleep
//
//	description:	Resume from sleep if abort_flag set (sec.)
//
//-------------------------------------------------------------------
void MenloLFC::abort_sleep(double time) 
{		
	
	for (int i = 0; i < (time/0.1); i++) {
		if (abortflag)
			break;
		omni_thread::sleep(0,100000000);
	}

}

//+------------------------------------------------------------------
/**
 *	method:	MenloLFC::string_explode
 * Explode a string 
*/
//+------------------------------------------------------------------
void MenloLFC::string_explode(string str, string separator, vector<unsigned int>* results, bool b_hex)
{
	int found;
	
	found = str.find_first_of(separator);
	while(found != string::npos) {
		if(found > 0) {
			unsigned int x;
			bool err = false;
			if(!b_hex)
			{
				std::istringstream i(str.substr(0,found));
				if (!(i >> x))
				{
					err = true;
				}
			}
			else
			{
				std::stringstream i; 
				i << hex << (str.substr(0,found));
				if (!(i >> x))
				{
					err = true;
				}
			}

			if (err)
			{
				DEBUG_STREAM << "MenloLFC::string_explode error converting '" << str << "'";
			}
			else
			{
				results->push_back(x);
				DEBUG_STREAM << "MenloLFC::string_explode pushing " << x << "=0x" << hex << x << dec;
			}
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		unsigned int x;
		bool err = false;
		if(!b_hex)
		{
			std::istringstream i(str);
			if (!(i >> x))
			{
				err = true;
			}
		}
		else
		{
			std::stringstream i; 
			i << hex << (str);
			if (!(i >> x))
			{
				err = true;
			}
		}
		if (err)
		{
			DEBUG_STREAM << "MenloLFC::string_explode error converting '" << str << "'";
		}
		else
		{
			results->push_back(x);
			DEBUG_STREAM << "MenloLFC::string_explode pushing " << x << "=0x" << hex << x << dec;
		}
	}

}

void MenloLFC::reconnect()
{
#if 0
	if(moxa)
	{
		/* Establish the connection to the server */
		if (connect(sock, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0)
		{
			stringstream tmp;
			tmp << "Connection failed to " << servIP  << ":" << servPort << ends;
			set_state(Tango::FAULT);
			set_status(tmp.str());
			conn_error = true;
			INFO_STREAM << "MenloLFC::reconnect(): " << tmp.str();
			return;
		}
		
	}
	else
	{	
		/* check if serialdevice is responding */
		try {
			if (devserial->ping()) 
			{
				devserial->set_timeout_millis(serial_timeout);
				//set_state(Tango::ON);	
				set_status("Connected to serialdevice " + serialdevice);
				INFO_STREAM << "Connected to serialdevice " + serialdevice << endl;
			}	
		} 
		catch(Tango::DevFailed e) {
			set_state(Tango::FAULT);
			Tango::DevState st = Tango::FAULT;
			push_change_event("state", &st);
			set_status("Failed to connect to serialdevice " + serialdevice);
			INFO_STREAM << "MenloLFC::reconnect Failed to connect to serialdevice " + serialdevice << endl; 
			conn_error = true;
			return;	
		}
		
		Tango::DeviceData Din;
		Tango::DevLong flush = 2;  //flush input and output	
		Din << flush;
		try	{
			devserial->command_inout("DevSerFlush",Din);
		}
		catch(Tango::DevFailed e) {
			string temp;
			temp = e.errors[0].desc;
			WARN_STREAM << "MenloLFC::reconnect error flushing serial device: " << temp << endl;
		}
	}
#endif	
	vector<string> names;
	string reg_s;
	data_t rdata;
	names.push_back(S_DEV);
	names.push_back(S_ADDR);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, S_DEV, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() == 0))
	{
		INFO_STREAM << "MenloTheta::reconnect: error reading Device Address, err=" << rdata.msg << endl;
		set_status("error reading Device Address, err=" + rdata.msg);
	}
	else
	{
		dev_add = rdata.data[0].front().u8;		//update device address with the one read from the device
		dev_add_loc = dev_add + 1;		//define local address as device address + 1
	}
	rdata.data[0].clear();
	abort_sleep(wr_sleep/1000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_ID);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, S_ID, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() == 0))
	{
		INFO_STREAM << "MenloTheta::reconnect: error reading Device ID, err=" << rdata.msg << endl;
		set_status("error reading Device ID, err=" + rdata.msg);
		strcpy(dev_id, "");
	}
	else
	{
		strcpy(dev_id, rdata.data[0].front().cstring);
	}
	
	rdata.data[0].clear();
	abort_sleep(wr_sleep/1000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_VER_HW);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, S_VER_HW, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 4))
	{
		INFO_STREAM << "MenloTheta::reconnect: error reading Device HW Version, err=" << rdata.msg << endl;
		set_status("error reading Device HW Version, err=" + rdata.msg);
		strcpy(dev_hw_ver, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int major,minor,patchlevel,build;
		deque<data_union>::iterator it = rdata.data[0].begin();
		build = it->u8;
		it++;
		patchlevel = it->u8;
		it++;
		minor = it->u8;
		it++;
		major = it->u8;
		sprintf(dev_hw_ver,"%d,%d,%d,%d", major,minor,patchlevel,build);
	}
	rdata.data[0].clear();
	abort_sleep(wr_sleep/1000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_VER_FW);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, S_VER_FW, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 4))
	{
		INFO_STREAM << "MenloTheta::reconnect: error reading Device FW Version, err=" << rdata.msg << endl;
		set_status("error reading Device FW Version, err=" + rdata.msg);
		strcpy(dev_fw_ver, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int major,minor,patchlevel,build;
		deque<data_union>::iterator it = rdata.data[0].begin();
		build = it->u8;
		it++;
		patchlevel = it->u8;
		it++;
		minor = it->u8;
		it++;
		major = it->u8;
		sprintf(dev_fw_ver,"%d,%d,%d,%d", major,minor,patchlevel,build);
	}
	rdata.data[0].clear();
	abort_sleep(wr_sleep/1000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_VER_DISPFW);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, S_VER_DISPFW, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 4))
	{
		INFO_STREAM << "MenloTheta::reconnect: error reading Display FW Version, err=" << rdata.msg << endl;
		set_status("error reading Display FW Version, err=" + rdata.msg);
		strcpy(display_fw_ver, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int major,minor,patchlevel,build;
		deque<data_union>::iterator it = rdata.data[0].begin();
		build = it->u8;
		it++;
		patchlevel = it->u8;
		it++;
		minor = it->u8;
		it++;
		major = it->u8;
		sprintf(display_fw_ver,"%d,%d,%d,%d", major,minor,patchlevel,build);
	}
	rdata.data[0].clear();
	abort_sleep(wr_sleep/1000);
	names.clear();
	names.push_back(S_DEV);
	names.push_back(S_SER);
	menlo->NamesCat(reg_s, names);
	WriteRead(PROTO_CMD_READ, S_SER, -1, rdata);
	if((rdata.msg != "") || (rdata.data[0].size() < 3))
	{
		INFO_STREAM << "MenloTheta::reconnect: error reading Device Serial, err=" << rdata.msg << endl;
		set_status("error reading Device Serial, err=" + rdata.msg);
		strcpy(dev_serial, "");
	}
	else
	{
		//strcpy(dev_id, rdata.data.front().cstring);
		int year,month,serial;
		deque<data_union>::iterator it = rdata.data[0].begin();
		year = it->u8;
		it++;
		month = it->u8;
		it++;
		serial = it->u16;
		sprintf(dev_serial,"%d,%d,%d", year,month,serial);
	}	

}

void MenloLFC::dump_msg(string label, int cnt, char *buf)
{
	stringstream tmp;
	for(int k=0;k<cnt;)
	{
		stringstream tmp;
		int g;
		tmp << label << ": "<<setw(3)<<setfill(' ')<<k<<" -> ";
		for(g=0; g<10; g++)
		{
			if(k+g>=cnt)
				break;
			tmp << " " <<hex<<setw(2)<<setfill(' ')<< (int)(unsigned char)buf[k+g];
		}
		k+=g;
		if(k<cnt)
		{
			tmp << " |  "<<dec<<setw(3)<<setfill(' ')<<k<<hex<<" -> ";
			for(g=0; g<10; g++)
			{
				if(k+g>=cnt)
					break;
				tmp << " " <<hex<<setw(2)<<setfill(' ')<< (int)(unsigned char)buf[k+g];
			}
			k+=g;
		}
		DEBUG_STREAM<<tmp.str()<<dec<<endl;
	}
	
}



}	//	namespace
