//
// file :         readthread.cpp
//
// description :  reading thread
//
// project :      TANGO Device Server
//
// Elettra-Sincrotrone Trieste S.C.p.A. di interesse nazionale
// Strada Statale 14 - km 163,5 in AREA Science Park
// 34149 Basovizza, Trieste ITALY
//

#include "readthread.h"
static const char __FILE__rev[] = __FILE__ " $Revision: 1.2 $";

//+------------------------------------------------------------------
//
//	method:			readthread::readthread()
//
//	description:	readthread constructor
//
//-------------------------------------------------------------------
readthread::readthread(MenloLFC_ns::MenloLFC *s):Tango::LogAdapter(s)
{ 
	DEBUG_STREAM << "readthread::readthread(): constructor... :" << __FILE__rev << endl;
	mydevice = s;
	//readSched = NULL;
	sched_mutex = new omni_mutex();
}

//+------------------------------------------------------------------
//
//	method:			readthread::~readthread()
//
//	description:	readthread destructor
//
//-------------------------------------------------------------------
readthread::~readthread()
{
	DEBUG_STREAM << "readthread::~readthread(): destructor... !" << endl;
	/*if(readSched)
	{
		delete readSched;
		readSched = NULL;
	}*/
	delete sched_mutex;
}

//+------------------------------------------------------------------
//
//	method:			readthread::run()
//
//	description:	Run
//
//-------------------------------------------------------------------
void readthread::run(void *) 
{
	INFO_STREAM << "readthread::run(): running... !" << endl;	
	int pausesec,pausenano,pausesec_dfl,pausenano_dfl,
	    cnt_err = 0,
		max_cnt_err = 3,
		ret;
	
	string reg;
	update_scheduler();
	INFO_STREAM << "readthread::run(): after update_scheduler" << endl;
	
	pausesec = pausesec_dfl = (int)(mydevice->th_sleep/1000);
	pausenano = pausenano_dfl = (int)(mydevice->th_sleep-pausesec*1000)*1000000;

	data_map::iterator it;// = mydevice->read_data.begin();
	//data_map::iterator it = mydevice->read_data.begin();

	while (!mydevice->abortflag) 
	{	
		//mydevice->mutex->lock();
		sched_mutex->lock();			
		//reg = readSched->next();
		reg = readSched.next();
		sched_mutex->unlock();
		it = mydevice->read_data.find(reg);
		/*reg = it->first*/
		ret = mydevice->WriteRead(PROTO_CMD_READ, reg, -1, it->second);
		//mydevice->mutex->unlock();
		
		/*it++;
		if(it == mydevice->read_data.end())
			it = mydevice->read_data.begin();*/
		if (ret < 0) 
			cnt_err++;
		else
			cnt_err = 0;	
		if (cnt_err > max_cnt_err) {
			pausesec = 5;
			pausenano = 0;
		}
		else {
			pausesec = pausesec_dfl;
			pausenano = pausenano_dfl;
		}
		//omni_thread::sleep(pausesec,pausenano);
		abort_sleep(mydevice->th_sleep/1000);		
	}
	INFO_STREAM << "readthread::run(): exit!!!" << endl;
}

//+------------------------------------------------------------------
//
//	method:			readthread::update_scheduler()
//
//	description:	update scheduler after new sensor has been detected
//
//-------------------------------------------------------------------
void readthread::update_scheduler()
{
	INFO_STREAM << "readthread::update_scheduler(): entering..." << endl;
	scheduler<string>::sched_t hi, low;
	sched_mutex->lock();
	/*if(readSched)
	{
		delete readSched;
		readSched = NULL;
	}*/
	for(data_map::iterator it2 = mydevice->read_data.begin(); it2 != mydevice->read_data.end(); it2++)
	{
		if(it2->second.pri == Hi_pri)
			hi.push_back(it2->first);
		else if(it2->second.pri == Low_pri)
			low.push_back(it2->first);			
	}
	//readSched = new scheduler<string>(hi, low, PRIORITY_FACTOR);
	readSched.update(hi, low, PRIORITY_FACTOR);
	sched_mutex->unlock();	
}

//+------------------------------------------------------------------
//
//	method:			readthread::force_scheduler()
//
//	description:	force scheduler to execute id as the first between
//					that with its priority
//
//-------------------------------------------------------------------
void readthread::force_scheduler(string id)
{
	sched_mutex->lock();
	//readSched->force(id);
	readSched.force(id);
	sched_mutex->unlock();
}

//+------------------------------------------------------------------
//
//	method:			readthread::abort_sleep
//
//	description:	Resume from sleep if abort_flag set (sec.)
//
//-------------------------------------------------------------------
void readthread::abort_sleep(double time)
{		
	
/*	for (int i = 0; i < (time/0.1); i++) {
		if (mydevice->abortflag)
			break;
		omni_thread::sleep(0,100000000);
	}*/
	usleep(long(time*1000000));

}

